import dataclasses
import functools
import logging
import os
import threading
import uuid
from collections import Counter
from inspect import signature
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import zmq
import zmq.asyncio

from scaler.client.agent.client_agent import ClientAgent
from scaler.client.agent.future_manager import ClientFutureManager
from scaler.client.future import ScalerFuture
from scaler.client.object_buffer import ObjectBuffer
from scaler.client.object_reference import ObjectReference
from scaler.client.serializer.default import DefaultSerializer
from scaler.client.serializer.mixins import Serializer
from scaler.io.config import DEFAULT_CLIENT_TIMEOUT_SECONDS, DEFAULT_HEARTBEAT_INTERVAL_SECONDS
from scaler.io.sync_connector import SyncConnector
from scaler.protocol.python.message import ClientDisconnect, ClientShutdownResponse, GraphTask, Task
from scaler.utility.exceptions import ClientQuitException
from scaler.utility.graph.optimization import cull_graph
from scaler.utility.graph.topological_sorter import TopologicalSorter
from scaler.utility.metadata.profile_result import ProfileResult
from scaler.utility.metadata.task_flags import TaskFlags, retrieve_task_flags_from_task
from scaler.utility.zmq_config import ZMQConfig, ZMQType
from scaler.worker.agent.processor.processor import Processor


@dataclasses.dataclass
class _CallNode:
    func: Callable
    args: Tuple[str, ...]

    def __post_init__(self):
        if not callable(self.func):
            raise TypeError(f"the first item of the tuple must be function, get {self.func}")

        if not isinstance(self.args, tuple):
            raise TypeError(f"arguments must be tuple, get {self.args}")

        for arg in self.args:
            if not isinstance(arg, str):
                raise TypeError(f"argument `{arg}` must be a string and the string has to be in the graph")


class Client:
    def __init__(
        self,
        address: str,
        profiling: bool = False,
        timeout_seconds: int = DEFAULT_CLIENT_TIMEOUT_SECONDS,
        heartbeat_interval_seconds: int = DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
        serializer: Serializer = DefaultSerializer(),
    ):
        """
        The Scaler Client used to send tasks to a scheduler.

        :param address: Address of Scheduler to submit work to
        :type address: str
        :param profiling: If True, the returned futures will have the `task_duration()` property enabled.
        :type profiling: bool
        :param timeout_seconds: Seconds until heartbeat times out
        :type timeout_seconds: int
        :param heartbeat_interval_seconds: Frequency of heartbeat to scheduler in seconds
        :type heartbeat_interval_seconds: int
        """
        self.__initialize__(address, profiling, timeout_seconds, heartbeat_interval_seconds, serializer)

    def __initialize__(
        self,
        address: str,
        profiling: bool,
        timeout_seconds: int,
        heartbeat_interval_seconds: int,
        serializer: Serializer = DefaultSerializer(),
    ):
        self._serializer = serializer

        self._profiling = profiling
        self._identity = f"{os.getpid()}|Client|{uuid.uuid4().bytes.hex()}".encode()

        self._client_agent_address = ZMQConfig(ZMQType.inproc, host=f"scaler_client_{uuid.uuid4().hex}")
        self._scheduler_address = ZMQConfig.from_string(address)
        self._timeout_seconds = timeout_seconds
        self._heartbeat_interval_seconds = heartbeat_interval_seconds

        self._stop_event = threading.Event()
        self._context = zmq.Context()
        self._connector = SyncConnector(
            context=self._context, socket_type=zmq.PAIR, address=self._client_agent_address, identity=self._identity
        )

        self._future_manager = ClientFutureManager(self._serializer)
        self._agent = ClientAgent(
            identity=self._identity,
            client_agent_address=self._client_agent_address,
            scheduler_address=ZMQConfig.from_string(address),
            context=self._context,
            future_manager=self._future_manager,
            stop_event=self._stop_event,
            timeout_seconds=self._timeout_seconds,
            heartbeat_interval_seconds=self._heartbeat_interval_seconds,
            serializer=self._serializer,
        )
        self._agent.start()

        logging.info(f"ScalerClient: connect to {self._scheduler_address.to_address()}")

        self._object_buffer = ObjectBuffer(self._identity, self._serializer, self._connector)
        self._future_factory = functools.partial(ScalerFuture, connector=self._connector)

        self._object_buffer.buffer_send_serializer()
        self._object_buffer.commit_send_objects()

    @property
    def identity(self):
        return self._identity

    def __del__(self):
        self.disconnect()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def __getstate__(self) -> dict:
        """
        Serializes the client object's state.

        Client serialization is useful when a client reference is used within a remote task:


        .. code:: python

            client = Client(...)

            def fibonacci(client: Client, n: int):
                if n == 0:
                    return 0
                elif n == 1:
                    return 1
                else:
                    a = client.submit(fibonacci, n - 1)
                    b = client.submit(fibonacci, n - 2)
                    return a.result() + b.result()

            print(client.submit(fibonacci, client, 7).result())


        When serializing the client, only saves the address parameters. When deserialized, a new client object
        connecting to the same scheduler and remote logger will be instantiated.
        """

        return {
            "address": self._scheduler_address.to_address(),
            "profiling": self._profiling,
            "timeout_seconds": self._timeout_seconds,
            "heartbeat_interval_seconds": self._heartbeat_interval_seconds,
        }

    def __setstate__(self, state: dict) -> None:
        # TODO: fix copy the serializer
        self.__initialize__(
            address=state["address"],
            profiling=state["profiling"],
            timeout_seconds=state["timeout_seconds"],
            heartbeat_interval_seconds=state["heartbeat_interval_seconds"],
        )

    def submit(self, fn: Callable, *args, **kwargs) -> ScalerFuture:
        """
        Submit a single task (function with arguments) to the scheduler, and return a future

        :param fn: function to be executed remotely
        :type fn: Callable
        :param args: positional arguments will be passed to function
        :return: future of the submitted task
        :rtype: ScalerFuture
        """

        self.__assert_client_not_stopped()

        function_object_id = self._object_buffer.buffer_send_function(fn).object_id
        all_args = Client.__convert_kwargs_to_args(fn, args, kwargs)

        task, future = self.__submit(function_object_id, all_args, delayed=True)

        self._object_buffer.commit_send_objects()
        self._connector.send(task)
        return future

    def map(self, fn: Callable, iterable: Iterable[Tuple[Any, ...]]) -> List[Any]:
        if not all(isinstance(args, (tuple, list)) for args in iterable):
            raise TypeError("iterable should be list of arguments(list or tuple-like) of function")

        self.__assert_client_not_stopped()

        function_object_id = self._object_buffer.buffer_send_function(fn).object_id
        tasks, futures = zip(*[self.__submit(function_object_id, args, delayed=False) for args in iterable])

        self._object_buffer.commit_send_objects()
        for task in tasks:
            self._connector.send(task)

        try:
            results = [fut.result() for fut in futures]
        except Exception as e:
            logging.exception(f"error happened when do scaler client.map:\n{e}")
            self.disconnect()
            raise e

        return results

    def get(
        self, graph: Dict[str, Union[Any, Tuple[Union[Callable, str], ...]]], keys: List[str], block: bool = True
    ) -> Dict[str, Union[Any, ScalerFuture]]:
        """
        .. code-block:: python
           :linenos:
            graph = {
                "a": 1,
                "b": 2,
                "c": (inc, "a"),
                "d": (inc, "b"),
                "e": (add, "c", "d")
            }

        :param graph: dictionary presentation of task graphs
        :type graph: Dict[str, Union[Any, Tuple[Union[Callable, Any]]
        :param keys: list of keys want to get results from computed graph
        :type keys: List[str]
        :param block: if True, it will directly return a dictionary that maps from keys to results
        :return: dictionary of mapping keys to futures, or map to results if block=True is specified
        :rtype: Dict[ScalerFuture]
        """

        self.__assert_client_not_stopped()

        graph = cull_graph(graph, keys)

        node_name_to_argument, call_graph = self.__split_data_and_graph(graph)
        self.__check_graph(node_name_to_argument, call_graph, keys)

        graph_task, compute_futures, finished_futures = self.__construct_graph(
            node_name_to_argument, call_graph, keys, block
        )
        self._object_buffer.commit_send_objects()
        self._connector.send(graph_task)

        self._future_manager.add_future(
            self._future_factory(
                task=Task.new_msg(
                    task_id=graph_task.task_id,
                    source=self._identity,
                    metadata=b"",
                    func_object_id=b"",
                    function_args=[],
                ),
                is_delayed=not block,
                group_task_id=None,
            )
        )
        for future in compute_futures.values():
            self._future_manager.add_future(future)

        # preserve the future insertion order based on inputted keys
        futures = {}
        for key in keys:
            if key in compute_futures:
                futures[key] = compute_futures[key]
            else:
                futures[key] = finished_futures[key]

        if not block:
            # just return futures
            return futures

        try:
            results = {k: v.result() for k, v in futures.items()}
        except Exception as e:
            logging.exception(f"error happened when do scaler client.get:\n{e}")
            self.disconnect()
            raise e

        return results

    def send_object(self, obj: Any, name: Optional[str] = None) -> ObjectReference:
        """
        send object to scheduler, this can be used to cache very large data to scheduler, and reuse it in multiple
        tasks

        :param obj: object to send, it will be serialized and send to scheduler
        :type obj: Any
        :param name: give a name to the cached argument
        :type name: Optional[str]
        :return: object reference
        :rtype ObjectReference
        """

        self.__assert_client_not_stopped()

        cache = self._object_buffer.buffer_send_object(obj, name)
        return ObjectReference(cache.object_name, cache.object_id, sum(map(len, cache.object_bytes)))

    def clear(self):
        """
        clear all resources used by the client, this will cancel all running futures and invalidate all existing object
        references
        """

        self._future_manager.cancel_all_futures()
        self._object_buffer.clear()

    def disconnect(self):
        """
        disconnect from connected scheduler, this will not shut down the scheduler
        """

        if self._stop_event.is_set():
            self.__destroy()
            return

        logging.info(f"ScalerClient: disconnect from {self._scheduler_address.to_address()}")

        self._future_manager.cancel_all_futures()

        self._connector.send(ClientDisconnect.new_msg(ClientDisconnect.DisconnectType.Disconnect))

        self.__destroy()

    def __receive_shutdown_response(self):
        message: Optional[ClientShutdownResponse] = None
        while not isinstance(message, ClientShutdownResponse):
            message = self._connector.receive()

        if not message.accepted:
            raise ValueError("Scheduler is in protected mode. Can't shutdown")

    def shutdown(self):
        """
        shutdown all workers that connected to the scheduler this client connects to, it will cancel all other
        clients' ongoing tasks, please be aware shutdown might not success if scheduler is configured as protected mode,
        then it cannot shut down scheduler and the workers
        """

        if not self._agent.is_alive():
            self.__destroy()
            return

        logging.info(f"ScalerClient: request shutdown for {self._scheduler_address.to_address()}")

        self._future_manager.cancel_all_futures()

        self._connector.send(ClientDisconnect.new_msg(ClientDisconnect.DisconnectType.Shutdown))
        try:
            self.__receive_shutdown_response()
        finally:
            self.__destroy()

    def __submit(self, function_object_id: bytes, args: Tuple[Any, ...], delayed: bool) -> Tuple[Task, ScalerFuture]:
        task_id = uuid.uuid1().bytes

        object_ids = []
        for arg in args:
            if isinstance(arg, ObjectReference):
                object_ids.append(arg.object_id)
            else:
                object_ids.append(self._object_buffer.buffer_send_object(arg).object_id)

        task_flags_bytes = self.__get_task_flags().serialize()

        arguments = [Task.Argument(Task.Argument.ArgumentType.ObjectID, object_id) for object_id in object_ids]
        task = Task.new_msg(
            task_id=task_id,
            source=self._identity,
            metadata=task_flags_bytes,
            func_object_id=function_object_id,
            function_args=arguments,
        )

        future = self._future_factory(task=task, is_delayed=delayed, group_task_id=None)
        self._future_manager.add_future(future)
        return task, future

    @staticmethod
    def __convert_kwargs_to_args(fn: Callable, args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> Tuple[Any, ...]:
        all_params = [p for p in signature(fn).parameters.values()]

        params = [p for p in all_params if p.kind in {p.POSITIONAL_ONLY, p.POSITIONAL_OR_KEYWORD}]

        if len(args) >= len(params):
            return args

        number_of_required = len([p for p in params if p.default is p.empty])

        args_list = list(args)
        kwargs = kwargs.copy()
        kwargs.update({p.name: p.default for p in all_params if p.kind == p.KEYWORD_ONLY if p.default != p.empty})

        for p in params[len(args_list) : number_of_required]:
            try:
                args_list.append(kwargs.pop(p.name))
            except KeyError:
                missing = tuple(p.name for p in params[len(args_list) : number_of_required])
                raise TypeError(f"{fn} missing {len(missing)} arguments: {missing}")

        for p in params[len(args_list) :]:
            args_list.append(kwargs.pop(p.name, p.default))

        return tuple(args_list)

    def __split_data_and_graph(
        self, graph: Dict[str, Union[Any, Tuple[Union[Callable, str], ...]]]
    ) -> Tuple[Dict[str, Tuple[Task.Argument, Any]], Dict[str, _CallNode]]:
        call_graph = {}
        node_name_to_argument: Dict[str, Tuple[Task.Argument, Union[Any, Tuple[Union[Callable, Any], ...]]]] = dict()

        for node_name, node in graph.items():
            if isinstance(node, tuple) and len(node) > 0 and callable(node[0]):
                call_graph[node_name] = _CallNode(func=node[0], args=node[1:])  # type: ignore[arg-type]
                continue

            if isinstance(node, ObjectReference):
                object_id = node.object_id
            else:
                object_id = self._object_buffer.buffer_send_object(node, name=node_name).object_id

            node_name_to_argument[node_name] = (Task.Argument(Task.Argument.ArgumentType.ObjectID, object_id), node)

        return node_name_to_argument, call_graph

    @staticmethod
    def __check_graph(
        node_to_argument: Dict[str, Tuple[Task.Argument, Any]], call_graph: Dict[str, _CallNode], keys: List[str]
    ):
        duplicate_keys = [key for key, count in dict(Counter(keys)).items() if count > 1]
        if duplicate_keys:
            raise KeyError(f"duplicate key detected in argument keys: {duplicate_keys}")

        # sanity check graph
        for key in keys:
            if key not in call_graph and key not in node_to_argument:
                raise KeyError(f"key {key} has to be in graph")

        sorter: TopologicalSorter[str] = TopologicalSorter()
        for node_name, node in call_graph.items():
            for arg in node.args:
                if arg not in node_to_argument and arg not in call_graph:
                    raise KeyError(f"argument {arg} in node '{node_name}': {node} is not defined in graph")

            sorter.add(node_name, *node.args)

        # check cyclic dependencies
        sorter.prepare()

    def __construct_graph(
        self,
        node_name_to_arguments: Dict[str, Tuple[Task.Argument, Any]],
        call_graph: Dict[str, _CallNode],
        keys: List[str],
        block: bool,
    ) -> Tuple[GraphTask, Dict[str, ScalerFuture], Dict[str, ScalerFuture]]:
        graph_task_id = uuid.uuid1().bytes

        node_name_to_task_id: Dict[str, bytes] = {node_name: uuid.uuid1().bytes for node_name in call_graph.keys()}

        task_flags_bytes = self.__get_task_flags().serialize()

        task_id_to_tasks = dict()

        for node_name, node in call_graph.items():
            task_id = node_name_to_task_id[node_name]
            function_cache = self._object_buffer.buffer_send_function(node.func)

            arguments: List[Task.Argument] = []
            for arg in node.args:
                assert arg in call_graph or arg in node_name_to_arguments

                if arg in call_graph:
                    arguments.append(
                        Task.Argument(type=Task.Argument.ArgumentType.Task, data=node_name_to_task_id[arg])
                    )
                elif arg in node_name_to_arguments:
                    argument, _ = node_name_to_arguments[arg]
                    arguments.append(argument)
                else:
                    raise ValueError("Not possible")

            task_id_to_tasks[task_id] = Task.new_msg(
                task_id=task_id,
                source=self._identity,
                metadata=task_flags_bytes,
                func_object_id=function_cache.object_id,
                function_args=arguments,
            )

        result_task_ids = [node_name_to_task_id[key] for key in keys if key in call_graph]
        graph_task = GraphTask.new_msg(graph_task_id, self._identity, result_task_ids, list(task_id_to_tasks.values()))

        compute_futures = {}
        ready_futures = {}
        for key in keys:
            if key in call_graph:
                compute_futures[key] = self._future_factory(
                    task=task_id_to_tasks[node_name_to_task_id[key]], is_delayed=not block, group_task_id=graph_task_id
                )

            elif key in node_name_to_arguments:
                argument, data = node_name_to_arguments[key]
                future: ScalerFuture = self._future_factory(
                    task=Task.new_msg(
                        task_id=argument.data, source=self._identity, metadata=b"", func_object_id=b"", function_args=[]
                    ),
                    is_delayed=False,
                    group_task_id=graph_task_id,
                )
                future.set_result_ready(argument.data, ProfileResult())
                future.set_result(data)
                ready_futures[key] = future

            else:
                raise ValueError(f"cannot find {key=} in graph")

        return graph_task, compute_futures, ready_futures

    def __get_task_flags(self) -> TaskFlags:
        parent_task_priority = self.__get_parent_task_priority()

        if parent_task_priority is not None:
            task_priority = parent_task_priority + 1
        else:
            task_priority = 0

        return TaskFlags(profiling=self._profiling, priority=task_priority)

    def __assert_client_not_stopped(self):
        if self._stop_event.is_set():
            raise ClientQuitException("client is already stopped.")

    def __destroy(self):
        self._agent.join()
        self._context.destroy(linger=1)

    @staticmethod
    def __get_parent_task_priority() -> Optional[int]:
        """If the client is running inside a Scaler processor, returns the priority of the associated task."""

        current_processor = Processor.get_current_processor()

        if current_processor is None:
            return None

        current_task = current_processor.current_task()
        assert current_task is not None

        return retrieve_task_flags_from_task(current_task).priority



import contextlib
import logging
import multiprocessing
import os
import signal
import uuid
from contextvars import ContextVar, Token
from multiprocessing.synchronize import Event as EventType
from typing import Callable, List, Optional, Tuple

import tblib.pickling_support
import zmq

from scaler.io.config import DUMMY_CLIENT
from scaler.io.sync_connector import SyncConnector
from scaler.io.utility import chunk_to_list_of_bytes
from scaler.protocol.python.common import ObjectContent, TaskStatus
from scaler.protocol.python.message import (
    ObjectInstruction,
    ObjectRequest,
    ObjectResponse,
    ProcessorInitialized,
    Task,
    TaskResult,
)
from scaler.protocol.python.mixins import Message
from scaler.utility.exceptions import MissingObjects
from scaler.utility.logging.utility import setup_logger
from scaler.utility.object_utility import generate_object_id, generate_serializer_object_id, serialize_failure
from scaler.utility.zmq_config import ZMQConfig
from scaler.worker.agent.processor.object_cache import ObjectCache

SUSPEND_SIGNAL = "SIGUSR1"  # use str instead of a signal.Signal to not trigger an import error on unsupported systems.

_current_processor: ContextVar[Optional["Processor"]] = ContextVar("_current_processor", default=None)


class Processor(multiprocessing.get_context("spawn").Process):  # type: ignore
    def __init__(
        self,
        event_loop: str,
        address: ZMQConfig,
        resume_event: Optional[EventType],
        resumed_event: Optional[EventType],
        garbage_collect_interval_seconds: int,
        trim_memory_threshold_bytes: int,
        logging_paths: Tuple[str, ...],
        logging_level: str,
    ):
        multiprocessing.Process.__init__(self, name="Processor")

        self._event_loop = event_loop
        self._address = address

        self._resume_event = resume_event
        self._resumed_event = resumed_event

        self._garbage_collect_interval_seconds = garbage_collect_interval_seconds
        self._trim_memory_threshold_bytes = trim_memory_threshold_bytes
        self._logging_paths = logging_paths
        self._logging_level = logging_level

        # self._client_to_decorator = {}

        self._object_cache: Optional[ObjectCache] = None

        self._current_task: Optional[Task] = None

    def run(self) -> None:
        self.__initialize()
        self.__run_forever()

    @staticmethod
    def get_current_processor() -> Optional["Processor"]:
        """Returns the current Processor instance controlling the current process, if any."""
        return _current_processor.get()

    def current_task(self) -> Optional[Task]:
        return self._current_task

    def __initialize(self):
        # modify the logging path and add process id to the path
        logging_paths = [f"{path}-{os.getpid()}" for path in self._logging_paths if path != "/dev/stdout"]
        if "/dev/stdout" in self._logging_paths:
            logging_paths.append("/dev/stdout")

        setup_logger(log_paths=tuple(logging_paths), logging_level=self._logging_level)
        tblib.pickling_support.install()

        self._connector = SyncConnector(
            context=zmq.Context(), socket_type=zmq.DEALER, address=self._address, identity=None
        )

        self._object_cache = ObjectCache(
            garbage_collect_interval_seconds=self._garbage_collect_interval_seconds,
            trim_memory_threshold_bytes=self._trim_memory_threshold_bytes,
        )
        self._object_cache.start()

        self.__register_signals()

    def __register_signals(self):
        self.__register_signal("SIGTERM", self.__interrupt)

        if self._resume_event is not None:
            self.__register_signal(SUSPEND_SIGNAL, self.__suspend)

    def __interrupt(self, *args):
        self._connector.close()  # interrupts any blocking socket.

    def __suspend(self, *args):
        assert self._resume_event is not None
        assert self._resumed_event is not None

        self._resume_event.wait()  # stops any computation in the main thread until the event is triggered

        # Ensures the processor agent knows we stopped waiting on `_resume_event`, as to avoid re-entrant wait on the
        # event.
        self._resumed_event.set()

    def __run_forever(self):
        try:
            self._connector.send(ProcessorInitialized.new_msg())
            while True:
                message = self._connector.receive()
                if message is None:
                    continue

                self.__on_connector_receive(message)

        except zmq.error.ZMQError as e:
            if e.errno != zmq.ENOTSOCK:  # ignore if socket got closed
                raise

        except (KeyboardInterrupt, InterruptedError):
            pass

        except Exception as e:
            logging.exception(f"Processor[{self.pid}]: failed with unhandled exception:\n{(e)}")

        finally:
            self._object_cache.destroy()
            self._connector.close()

            self._object_cache.join()

    def __on_connector_receive(self, message: Message):
        if isinstance(message, ObjectInstruction):
            self.__on_receive_object_instruction(message)
            return

        if isinstance(message, ObjectResponse):
            self.__on_receive_object_response(message)
            return

        if isinstance(message, Task):
            self.__on_received_task(message)
            return

        logging.error(f"unknown {message=}")

    def __on_receive_object_instruction(self, instruction: ObjectInstruction):
        if instruction.instruction_type == ObjectInstruction.ObjectInstructionType.Delete:
            for object_id in instruction.object_content.object_ids:
                self._object_cache.del_object(object_id)
            return

        logging.error(f"worker received unknown object instruction type {instruction=}")

    def __on_receive_object_response(self, response: ObjectResponse):
        if response.response_type == ObjectResponse.ObjectResponseType.Content:
            self.__on_receive_object_content(response.object_content)
            return

        logging.error(f"worker received unknown request object request type {response=}")

    def __on_receive_object_content(self, object_content: ObjectContent):
        task = self._current_task
        assert task is not None

        self._object_cache.add_objects(object_content, task)

        # if still have unknown objects, something awful happened
        unknown_object_ids = self.__get_not_ready_object_ids(task)
        if not unknown_object_ids:
            self.__process_task(task)
            return

        logging.error(f"Task({task.task_id.hex()}): insufficient objects received from scheduler, task failed")
        self.__send_result(
            task.source,
            task.task_id,
            TaskStatus.Failed,
            serialize_failure(MissingObjects(",".join([obj.hex() for obj in unknown_object_ids]))),
        )

    def __on_received_task(self, task: Task):
        self._current_task = task

        unknown_object_ids = self.__get_not_ready_object_ids(task)
        if not unknown_object_ids:
            self.__process_task(task)
            return

        self._connector.send(ObjectRequest.new_msg(ObjectRequest.ObjectRequestType.Get, unknown_object_ids))

    def __get_not_ready_object_ids(self, task: Task) -> Tuple[bytes, ...]:
        required_object_ids = self.__get_required_object_ids_for_task(task)
        return tuple(filter(lambda oid: not self._object_cache.has_object(oid), required_object_ids))

    @staticmethod
    def __get_required_object_ids_for_task(task: Task) -> List[bytes]:
        serializer_id = generate_serializer_object_id(task.source)
        object_ids = [serializer_id, task.func_object_id]
        object_ids.extend(
            [argument.data for argument in task.function_args if argument.type == Task.Argument.ArgumentType.ObjectID]
        )
        return object_ids

    def __process_task(self, task: Task):
        try:
            function = self._object_cache.get_object(task.func_object_id)
            function_with_logger = self.__get_object_with_client_logger(DUMMY_CLIENT, function)

            args = [self._object_cache.get_object(arg.data) for arg in task.function_args]

            with self.__processor_context():
                result = function_with_logger(*args)

            result_bytes = self._object_cache.serialize(task.source, result)
            status = TaskStatus.Success

        except Exception as e:
            logging.exception(f"exception when processing task_id={task.task_id.hex()}:")
            status = TaskStatus.Failed
            result_bytes = serialize_failure(e)

        self.__send_result(task.source, task.task_id, status, result_bytes)

    def __get_object_with_client_logger(self, client: bytes, fn: Callable) -> Callable:
        assert self is not None
        assert isinstance(client, bytes)
        return fn

        # if client in self._client_to_decorator:
        #     wrap = self._client_to_decorator[client]
        #     return wrap(fn)
        #
        # def _generate_wrapper(handler: WorkerLogPublisher):
        #     def decorator(func: Callable):
        #         @functools.wraps(func)
        #         def wrapper(*args, **kwargs):
        #             logger = logging.getLogger()
        #             logger.addHandler(handler)
        #             result = func(*args, **kwargs)
        #             logger.removeHandler(handler)
        #             return result
        #
        #         return wrapper
        #
        #     return decorator
        #
        # wrap = _generate_wrapper(WorkerLogPublisher(client, self._log_hub_address, log_level=logging.DEBUG))
        # self._client_to_decorator[client] = wrap
        # return wrap(fn)

    def __send_result(self, source: bytes, task_id: bytes, status: TaskStatus, result_bytes: bytes):
        self._current_task = None

        # use uuid to avoid object_id collision, each result object_id is unique, even it's the same object across
        # clients
        result_object_id = generate_object_id(source, uuid.uuid4().bytes)
        self._connector.send(
            ObjectInstruction.new_msg(
                ObjectInstruction.ObjectInstructionType.Create,
                source,
                ObjectContent.new_msg(
                    (result_object_id,),
                    (ObjectContent.ObjectContentType.Object,),
                    (f"<res {result_object_id.hex()[:6]}>".encode(),),
                    (chunk_to_list_of_bytes(result_bytes),),
                ),
            )
        )
        self._connector.send(TaskResult.new_msg(task_id, status, metadata=b"", results=[result_object_id]))

    @staticmethod
    def __set_current_processor(context: Optional["Processor"]) -> Token:
        if context is not None and _current_processor.get() is not None:
            raise ValueError("cannot override a previously set processor context.")

        return _current_processor.set(context)

    @contextlib.contextmanager
    def __processor_context(self):
        self.__set_current_processor(self)
        try:
            yield
        finally:
            self.__set_current_processor(None)

    @staticmethod
    def __register_signal(signal_name: str, handler: Callable) -> None:
        signal_instance = getattr(signal, signal_name, None)
        if signal_instance is None:
            raise RuntimeError(f"unsupported platform, signal not availaible: {signal_name}.")

        signal.signal(signal_instance, handler)



import dataclasses
import logging
from asyncio import Queue
from typing import List, Optional, Set

from scaler.io.async_binder import AsyncBinder
from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.common import ObjectContent
from scaler.protocol.python.message import ObjectInstruction, ObjectRequest, ObjectResponse
from scaler.protocol.python.status import ObjectManagerStatus
from scaler.scheduler.mixins import ClientManager, ObjectManager, WorkerManager
from scaler.scheduler.object_usage.object_tracker import ObjectTracker, ObjectUsage
from scaler.utility.formatter import format_bytes
from scaler.utility.mixins import Looper, Reporter


@dataclasses.dataclass
class _ObjectCreation(ObjectUsage):
    object_id: bytes
    object_creator: bytes
    object_type: ObjectContent.ObjectContentType
    object_name: bytes
    object_bytes: List[bytes]

    def get_object_key(self) -> bytes:
        return self.object_id


class VanillaObjectManager(ObjectManager, Looper, Reporter):
    def __init__(self):
        self._object_storage: ObjectTracker[bytes, _ObjectCreation] = ObjectTracker(
            "object_usage", self.__finished_object_storage
        )

        self._queue_deleted_object_ids: Queue[bytes] = Queue()

        self._binder: Optional[AsyncBinder] = None
        self._binder_monitor: Optional[AsyncConnector] = None
        self._client_manager: Optional[ClientManager] = None
        self._worker_manager: Optional[WorkerManager] = None

    def register(
        self,
        binder: AsyncBinder,
        binder_monitor: AsyncConnector,
        client_manager: ClientManager,
        worker_manager: WorkerManager,
    ):
        self._binder = binder
        self._binder_monitor = binder_monitor
        self._client_manager = client_manager
        self._worker_manager = worker_manager

    async def on_object_instruction(self, source: bytes, instruction: ObjectInstruction):
        if instruction.instruction_type == ObjectInstruction.ObjectInstructionType.Create:
            self.__on_object_create(source, instruction)
            return

        if instruction.instruction_type == ObjectInstruction.ObjectInstructionType.Delete:
            self.on_del_objects(instruction.object_user, set(instruction.object_content.object_ids))
            return

        logging.error(
            f"received unknown object response type instruction_type={instruction.instruction_type} from "
            f"source={instruction.object_user!r}"
        )

    async def on_object_request(self, source: bytes, request: ObjectRequest):
        if request.request_type == ObjectRequest.ObjectRequestType.Get:
            await self.__process_get_request(source, request)
            return

        logging.error(f"received unknown object request type {request=} from {source=!r}")

    def on_add_object(
        self,
        object_user: bytes,
        object_id: bytes,
        object_type: ObjectContent.ObjectContentType,
        object_name: bytes,
        object_bytes: List[bytes],
    ):
        creation = _ObjectCreation(object_id, object_user, object_type, object_name, object_bytes)
        logging.debug(
            f"add object cache "
            f"object_name={creation.object_name!r}, "
            f"object_type={creation.object_type}, "
            f"object_id={creation.object_id.hex()}, "
            f"size={format_bytes(len(creation.object_bytes))}"
        )

        self._object_storage.add_object(creation)
        self._object_storage.add_blocks_for_one_object(creation.get_object_key(), {creation.object_creator})

    def on_del_objects(self, object_user: bytes, object_ids: Set[bytes]):
        for object_id in object_ids:
            self._object_storage.remove_one_block_for_objects({object_id}, object_user)

    def clean_client(self, client: bytes):
        self._object_storage.remove_blocks({client})

    async def routine(self):
        await self.__routine_send_objects_deletions()

    def has_object(self, object_id: bytes) -> bool:
        return self._object_storage.has_object(object_id)

    def get_object_name(self, object_id: bytes) -> bytes:
        if not self.has_object(object_id):
            return b"<Unknown>"

        return self._object_storage.get_object(object_id).object_name

    def get_object_content(self, object_id: bytes) -> List[bytes]:
        if not self.has_object(object_id):
            return list()

        return self._object_storage.get_object(object_id).object_bytes

    def get_status(self) -> ObjectManagerStatus:
        return ObjectManagerStatus.new_msg(
            self._object_storage.object_count(),
            sum(sum(map(len, v.object_bytes)) for _, v in self._object_storage.items()),
        )

    async def __process_get_request(self, source: bytes, request: ObjectRequest):
        await self._binder.send(source, self.__construct_response(request))

    async def __routine_send_objects_deletions(self):
        deleted_object_ids = [await self._queue_deleted_object_ids.get()]
        self._queue_deleted_object_ids.task_done()

        while not self._queue_deleted_object_ids.empty():
            deleted_object_ids.append(self._queue_deleted_object_ids.get_nowait())
            self._queue_deleted_object_ids.task_done()

        for worker in self._worker_manager.get_worker_ids():
            await self._binder.send(
                worker,
                ObjectInstruction.new_msg(
                    ObjectInstruction.ObjectInstructionType.Delete,
                    worker,
                    ObjectContent.new_msg(tuple(deleted_object_ids)),
                ),
            )

    def __on_object_create(self, source: bytes, instruction: ObjectInstruction):
        if not self._client_manager.has_client_id(instruction.object_user):
            logging.error(f"received object creation from {source!r} for unknown client {instruction.object_user!r}")
            return

        for object_id, object_type, object_name, object_bytes in zip(
            instruction.object_content.object_ids,
            instruction.object_content.object_types,
            instruction.object_content.object_names,
            instruction.object_content.object_bytes,
        ):
            self.on_add_object(instruction.object_user, object_id, object_type, object_name, object_bytes)

    def __finished_object_storage(self, creation: _ObjectCreation):
        logging.debug(
            f"del object cache "
            f"object_name={creation.object_name!r}, "
            f"object_id={creation.object_id.hex()}, "
            f"size={format_bytes(len(creation.object_bytes))}"
        )
        self._queue_deleted_object_ids.put_nowait(creation.object_id)

    def __construct_response(self, request: ObjectRequest) -> ObjectResponse:
        object_ids = []
        object_types = []
        object_names = []
        object_bytes = []
        for object_id in request.object_ids:
            if not self.has_object(object_id):
                continue

            object_info = self._object_storage.get_object(object_id)
            object_ids.append(object_info.object_id)
            object_types.append(object_info.object_type)
            object_names.append(object_info.object_name)
            object_bytes.append(object_info.object_bytes)

        return ObjectResponse.new_msg(
            ObjectResponse.ObjectResponseType.Content,
            ObjectContent.new_msg(
                tuple(request.object_ids), tuple(object_types), tuple(object_names), tuple(object_bytes)
            ),
        )



import random

from dask.distributed import Client as DaskClient

from scaler.utility.logging.scoped_logger import ScopedLogger
from scaler.utility.logging.utility import setup_logger


def sleep_print(sec: int):
    return sec * 1


def main():
    setup_logger()

    tasks = [random.randint(0, 100) for _ in range(10000)]

    client = DaskClient("127.0.0.1:12345")

    with ScopedLogger(f"submit {len(tasks)} tasks"):
        futures = [client.submit(sleep_print, i) for i in tasks]

    with ScopedLogger(f"gather {len(futures)} results"):
        results = [future.result() for future in futures]

    assert results == tasks


if __name__ == "__main__":
    main()



import dataclasses
from typing import Dict, List, Optional

from nicegui import ui
from nicegui.element import Element

from scaler.protocol.python.status import ProcessorStatus, WorkerStatus
from scaler.ui.utility import format_worker_name


@dataclasses.dataclass
class WorkerProcessors:
    workers: Dict[bytes, "WorkerProcessorTable"] = dataclasses.field(default_factory=dict)

    _parent: Optional[Element] = dataclasses.field(default=None)

    def draw_section(self, parent: Element):
        self._parent = parent
        for processor_table in self.workers.values():
            processor_table.draw_table()

    def update_data(self, data: List[WorkerStatus]):
        assert self._parent is not None

        previous_worker_ids = set(self.workers.keys())

        with self._parent:
            for worker in data:
                worker_name = worker.worker_id.decode()

                processor_table = self.workers.get(worker.worker_id)

                if processor_table is None:
                    processor_table = WorkerProcessorTable(worker_name, worker.rss_free, worker.processor_statuses)
                    processor_table.draw_table()
                    self.workers[worker.worker_id] = processor_table
                elif processor_table.processor_statuses != worker.processor_statuses:
                    processor_table.processor_statuses = worker.processor_statuses
                    processor_table.draw_table.refresh()

            removed_workers = previous_worker_ids - set(worker.worker_id for worker in data)
            for worker_id in removed_workers:
                self.workers.pop(worker_id).delete_row()


@dataclasses.dataclass
class WorkerProcessorTable:
    worker_name: str
    rss_free: int
    processor_statuses: List[ProcessorStatus]

    handler: Optional[Element] = dataclasses.field(default=None)

    @ui.refreshable
    def draw_table(self):
        formatted_worker_name = format_worker_name(self.worker_name)
        with ui.card().classes("w-full") as handler:
            self.handler = handler

            ui.markdown(f"Worker **{formatted_worker_name}**").classes("text-xl")

            with ui.grid(columns=6).classes("w-full"):
                self.draw_titles()
                for processor in sorted(self.processor_statuses, key=lambda x: x.pid):
                    self.draw_row(processor, self.rss_free)

    @staticmethod
    def draw_titles():
        ui.label("Processor PID")
        ui.label("CPU %")
        ui.label("RSS (in MB)")
        ui.label("Initialized")
        ui.label("Has Task")
        ui.label("Suspended")

    @staticmethod
    def draw_row(processor_status: ProcessorStatus, rss_free: int):
        cpu = processor_status.resource.cpu
        rss = int(processor_status.resource.rss / 1e6)
        rss_free = int(rss_free / 1e6)

        ui.label(processor_status.pid)
        ui.knob(value=cpu, track_color="grey-2", show_value=True, min=0, max=100)
        ui.knob(value=rss, track_color="grey-2", show_value=True, min=0, max=rss + rss_free)
        ui.checkbox().bind_value_from(processor_status, "initialized")
        ui.checkbox().bind_value_from(processor_status, "has_task")
        ui.checkbox().bind_value_from(processor_status, "suspended")

    def delete_row(self):
        assert self.handler is not None
        self.handler.clear()
        self.handler.delete()



import dataclasses

from scaler.utility.zmq_config import ZMQConfig


@dataclasses.dataclass
class SchedulerConfig:
    event_loop: str
    address: ZMQConfig
    io_threads: int
    max_number_of_tasks_waiting: int
    per_worker_queue_size: int
    client_timeout_seconds: int
    worker_timeout_seconds: int
    object_retention_seconds: int
    load_balance_seconds: int
    load_balance_trigger_times: int
    protected: bool



import unittest

from tests.utility import logging_test_name

from scaler.protocol.python.common import ObjectContent
from scaler.protocol.python.message import ObjectInstruction, ObjectRequest, ObjectResponse
from scaler.utility.logging.utility import setup_logger
from scaler.worker.agent.object_tracker import VanillaObjectTracker


class TestWorkerObjectTracker(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)

    def test_object_tracker(self) -> None:
        tracker = VanillaObjectTracker()

        tracker.on_object_request(
            b"processor_1", ObjectRequest.new_msg(ObjectRequest.ObjectRequestType.Get, (b"object_1", b"object_2"))
        )

        tracker.on_object_request(
            b"processor_2", ObjectRequest.new_msg(ObjectRequest.ObjectRequestType.Get, (b"object_1", b"object_2"))
        )
        tracker.on_object_request(
            b"processor_2", ObjectRequest.new_msg(ObjectRequest.ObjectRequestType.Get, (b"object_3",))
        )

        tracker.on_object_request(
            b"processor_3", ObjectRequest.new_msg(ObjectRequest.ObjectRequestType.Get, (b"object_4", b"object_5"))
        )

        response_1 = tracker.on_object_response(
            ObjectResponse.new_msg(
                ObjectResponse.ObjectResponseType.Content, ObjectContent.new_msg((b"object_1", b"object_2"))
            )
        )
        self.assertSetEqual(set(response_1), {b"processor_1", b"processor_2"})

        response_2 = tracker.on_object_response(
            ObjectResponse.new_msg(
                ObjectResponse.ObjectResponseType.Content, ObjectContent.new_msg((b"object_unknown",))
            )
        )
        self.assertSetEqual(response_2, set())

        response_3 = tracker.on_object_response(
            ObjectResponse.new_msg(ObjectResponse.ObjectResponseType.Content, ObjectContent.new_msg((b"object_3",)))
        )
        self.assertSetEqual(response_3, {b"processor_2"})

        object_instructions = tracker.on_object_instruction(
            ObjectInstruction.new_msg(
                ObjectInstruction.ObjectInstructionType.Delete,
                b"client",
                ObjectContent.new_msg(
                    (b"object_1", b"object_2", b"object_3"),
                    tuple([ObjectContent.ObjectContentType.Object] * 3),
                    (b"name_1", b"name_2", b"name_3"),
                    ([b"content_1"], [b"content_2"], [b"content_3"]),
                ),
            )
        )
        self.assertSetEqual(set(object_instructions.keys()), {b"processor_1", b"processor_2"})

        tracker.on_processor_end(b"processor_3")

        self.assertEqual(len(tracker._object_request_to_processor_ids.left_keys()), 0)
        self.assertEqual(len(tracker._object_id_to_processors_ids.left_keys()), 0)



import abc
from typing import TypeVar


class Message(metaclass=abc.ABCMeta):
    def __init__(self, msg):
        self._msg = msg

    def get_message(self):
        return self._msg


MessageType = TypeVar("MessageType", bound=Message)



import datetime
from queue import SimpleQueue
from threading import Lock
from typing import Dict, List, Optional, Set, Tuple

from nicegui import ui

from scaler.protocol.python.common import TaskStatus
from scaler.protocol.python.message import StateTask
from scaler.ui.live_display import WorkersSection
from scaler.ui.setting_page import Settings
from scaler.ui.utility import format_timediff, format_worker_name, get_bounds, make_tick_text, make_ticks


class TaskColors:
    ONGOING = "yellow"
    NOWORK = "white"
    SUCCESS = "green"
    FAILED = "red"
    UNKNOWN = "lightgray"
    DEADWORKER = NOWORK
    CANCELED = "purple"
    CANCELING = CANCELED

    __task_status_to_color = {
        TaskStatus.Success: SUCCESS,
        TaskStatus.Failed: FAILED,
        TaskStatus.Canceled: CANCELED,
        TaskStatus.Canceling: CANCELING,
    }

    @staticmethod
    def from_status(status: TaskStatus) -> str:
        return TaskColors.__task_status_to_color[status]


class TaskStream:
    def __init__(self):
        self._figure = {}
        self._plot = None

        self._settings: Optional[Settings] = None

        self._start_time = datetime.datetime.now() - datetime.timedelta(minutes=30)
        self._last_task_tick = datetime.datetime.now()

        self._current_tasks: Dict[str, Tuple[bool, Set[bytes], Optional[datetime.datetime]]] = {}
        self._completed_data_cache: Dict[str, Dict] = {}

        self._worker_to_object_name: Dict[str, str] = {}
        self._worker_last_update: Dict[str, datetime.datetime] = {}
        self._task_id_to_worker: Dict[bytes, str] = {}

        self._seen_workers = set()
        self._lost_workers_queue: SimpleQueue[Tuple[datetime.datetime, str]] = SimpleQueue()

        self._data_update_lock = Lock()
        self._busy_workers: Set[str] = set()
        self._busy_workers_update_time: datetime.datetime = datetime.datetime.now()

    def setup_task_stream(self, settings: Settings):
        with ui.card().classes("w-full").style("height: 85vh"):
            fig = {
                "data": [],
                "layout": {
                    "barmode": "stack",
                    "autosize": True,
                    "margin": {"l": 163},
                    "xaxis": {
                        "autorange": False,
                        "range": [0, 300],
                        "showgrid": False,
                        "tickmode": "array",
                        "tickvals": [0, 50, 100, 150, 200, 250, 300],
                        "ticktext": [-300, -250, -200, -150, -100, -50, 0],
                        "zeroline": False,
                    },
                    "yaxis": {
                        "autorange": True,
                        "automargin": True,
                        "rangemode": "nonnegative",
                        "showgrid": False,
                        "type": "category",
                    },
                },
            }
            self._figure = fig
            self._completed_data_cache = {}
            self._plot = ui.plotly(self._figure).classes("w-full h-full")
            self._settings = settings

    def __setup_worker_cache(self, worker: str):
        if worker in self._completed_data_cache:
            return
        self._completed_data_cache[worker] = {
            "type": "bar",
            "name": "History",
            "y": [],
            "x": [],
            "orientation": "h",
            "marker": {"color": [], "width": 5},
            "hovertemplate": [],
            "hovertext": [],
            "showlegend": False,
        }

    def __get_history_fields(self, worker: str, index: int) -> Tuple[float, str, str]:
        worker_data = self._completed_data_cache[worker]
        time_taken = worker_data["x"][index]
        color = worker_data["marker"]["color"][index]
        text = worker_data["hovertext"][index]
        return time_taken, color, text

    def __remove_last_elements(self, worker: str):
        worker_data = self._completed_data_cache[worker]
        del worker_data["y"][-1]
        del worker_data["x"][-1]
        del worker_data["marker"]["color"][-1]
        del worker_data["hovertext"][-1]
        del worker_data["hovertemplate"][-1]

    def __add_bar(self, worker: str, time_taken: float, task_color: str, hovertext: str):
        worker_history = self._completed_data_cache[worker]
        if len(worker_history["y"]) > 1:
            last_time_taken, last_color, last_text = self.__get_history_fields(worker, -1)

            # lengthen last bar if they're the same type
            if last_color == task_color and last_text == hovertext:
                worker_history["x"][-1] += time_taken
                return

            # if there's a short gap from last task to current task, merge the bars
            # this serves two purposes:
            #   - get a clean bar instead of many ~0 width lines
            #   - more importantly, make the ui significantly more responsive
            if task_color != TaskColors.NOWORK and len(worker_history["y"]) > 2:
                penult_time_taken, penult_color, penult_text = self.__get_history_fields(worker, -2)

                if last_time_taken < 0.1 and penult_color == task_color and penult_text == hovertext:
                    worker_history["x"][-2] += time_taken + last_time_taken
                    self.__remove_last_elements(worker)
                    return

        self._completed_data_cache[worker]["y"].append(format_worker_name(worker))
        self._completed_data_cache[worker]["x"].append(time_taken)
        self._completed_data_cache[worker]["marker"]["color"].append(task_color)
        self._completed_data_cache[worker]["hovertext"].append(hovertext)

        if hovertext:
            self._completed_data_cache[worker]["hovertemplate"].append("%{hovertext} (%{x})")
        else:
            self._completed_data_cache[worker]["hovertemplate"].append("")

    def __handle_task_result(self, state: StateTask, now: datetime.datetime):
        worker = self._task_id_to_worker.get(state.task_id, "")
        if worker == "":
            return

        task_status = state.status
        self._worker_last_update[worker] = now

        _, _, start = self._current_tasks.get(worker, (False, set(), None))

        if start is None:
            # we don't know when this task started, so just ignore
            return

        with self._data_update_lock:
            self.__remove_task_from_worker(worker=worker, task_id=state.task_id, now=now, force_new_time=True)
        self.__add_bar(
            worker,
            format_timediff(start, now),
            TaskColors.from_status(task_status),
            self._worker_to_object_name.get(worker, ""),
        )

    def __handle_new_worker(self, worker: str, now: datetime.datetime):
        if worker not in self._completed_data_cache:
            self.__setup_worker_cache(worker)
            self.__add_bar(worker, format_timediff(self._start_time, now), TaskColors.DEADWORKER, "")
        self._seen_workers.add(worker)

    def __remove_task_from_worker(self, worker: str, task_id: bytes, now: datetime.datetime, force_new_time: bool):
        doing_job, task_list, prev_start_time = self._current_tasks[worker]

        task_list.remove(task_id)

        self._current_tasks[worker] = (len(task_list) != 0, task_list, now if force_new_time else prev_start_time)

    def __handle_running_task(self, state: StateTask, worker: str, now: datetime.datetime):
        # if another worker was previously assigned this task, remove it
        previous_worker = self._task_id_to_worker.get(state.task_id)
        if previous_worker and previous_worker != worker:
            self.__remove_task_from_worker(worker=previous_worker, task_id=state.task_id, now=now, force_new_time=False)

        self._task_id_to_worker[state.task_id] = worker
        self._worker_to_object_name[worker] = state.function_name.decode()

        doing_job, task_list, start_time = self._current_tasks.get(worker, (False, set(), None))
        if doing_job:
            with self._data_update_lock:
                self._current_tasks[worker][1].add(state.task_id)
                return

        with self._data_update_lock:
            self._current_tasks[worker] = (True, {state.task_id}, now)
        if start_time:
            self.__add_bar(worker, format_timediff(start_time, now), TaskColors.NOWORK, "")

    def handle_task_state(self, state: StateTask):
        """
        The scheduler sends out `state.worker` while a Task is running.
        However, as soon as the task is done, that entry is cleared.
        A Success status will thus come with an empty `state.worker`, so
        we store this mapping ourselves based on the Running statuses we see.
        """

        task_status = state.status
        now = datetime.datetime.now()
        self._last_task_tick = now

        if task_status in {TaskStatus.Success, TaskStatus.Failed, TaskStatus.Canceling}:
            self.__handle_task_result(state, now)
            return

        if not (worker := state.worker):
            return

        worker_string = worker.decode()
        self._worker_last_update[worker_string] = now

        if worker_string not in self._seen_workers:
            self.__handle_new_worker(worker_string, now)

        if task_status in {TaskStatus.Running}:
            self.__handle_running_task(state, worker_string, now)

    def __add_lost_worker(self, worker: str, now: datetime.datetime):
        self._lost_workers_queue.put((now, worker))

    def __detect_lost_workers(self, now: datetime.datetime):
        removed_workers = []
        for worker in self._current_tasks.keys():
            last_tick = self._worker_last_update[worker]
            if now - last_tick > self._settings.memory_store_time:
                self.__add_lost_worker(worker, now)
                removed_workers.append(worker)

        for worker in removed_workers:
            del self._current_tasks[worker]

    def __remove_worker_from_history(self, worker: str):
        del self._completed_data_cache[worker]
        self._seen_workers.remove(worker)

    def __remove_old_workers(self, remove_up_to: datetime.datetime):
        while not self._lost_workers_queue.empty():
            timestamp, worker = self._lost_workers_queue.get()
            if timestamp > remove_up_to:
                self._lost_workers_queue.put((timestamp, worker))
                return
            self.__remove_worker_from_history(worker)

    def __split_workers_by_status(self, now: datetime.datetime) -> List[Tuple[str, float, str]]:
        workers_doing_jobs = []
        for worker, (doing_job, task_list, start_time) in self._current_tasks.items():
            if doing_job:
                worker_name = format_worker_name(worker)
                duration = format_timediff(start_time, now)
                object_name = self._worker_to_object_name.get(worker, "")

                workers_doing_jobs.append((worker_name, duration, object_name))
        return workers_doing_jobs

    def update_data(self, workers_section: WorkersSection):
        now = datetime.datetime.now()
        worker_names = sorted(workers_section.workers.keys())
        itls = {w: workers_section.workers[w].itl for w in worker_names}
        busy_workers = {w for w in worker_names if len(itls[w]) == 3 and itls[w][1] == "1" and itls[w][2] == "1"}
        for worker in worker_names:
            self._worker_last_update[worker] = now

        with self._data_update_lock:
            self._busy_workers = busy_workers
            self._busy_workers_update_time = now

    def clear_stale_busy_workers(self, now: datetime.datetime):
        if now - self._busy_workers_update_time > datetime.timedelta(seconds=2):
            self._busy_workers = set()

    def update_plot(self):
        with self._data_update_lock:
            now = datetime.datetime.now()

            self.clear_stale_busy_workers(now)

            task_update_time = self._last_task_tick
            workers_doing_tasks = self.__split_workers_by_status(now)

            self.__detect_lost_workers(now)
            remove_up_to = now - self._settings.memory_store_time
            self.__remove_old_workers(remove_up_to)

            completed_cache_values = list(self._completed_data_cache.values())

        if now - task_update_time >= datetime.timedelta(seconds=30):
            # get rid of the in-progress plots in ['data']
            self._figure["data"] = completed_cache_values
            self.__render_plot(now)
            return

        working_data = {
            "type": "bar",
            "name": "Working",
            "y": [w for (w, _, _) in workers_doing_tasks],
            "x": [t for (_, t, _) in workers_doing_tasks],
            "orientation": "h",
            "text": [f for (_, _, f) in workers_doing_tasks],
            "hovertemplate": "%{text} (%{x})",
            "marker": {"color": TaskColors.ONGOING, "width": 5},
            "showlegend": False,
        }
        plot_data = completed_cache_values + [working_data]
        self._figure["data"] = plot_data
        self.__render_plot(now)

    def __render_plot(self, now: datetime.datetime):
        lower_bound, upper_bound = get_bounds(now, self._start_time, self._settings)

        ticks = make_ticks(lower_bound, upper_bound)
        tick_text = make_tick_text(int(self._settings.stream_window.total_seconds()))

        self._figure["layout"]["xaxis"]["range"] = [lower_bound, upper_bound]
        self._figure["layout"]["xaxis"]["tickvals"] = ticks
        self._figure["layout"]["xaxis"]["ticktext"] = tick_text
        self._plot.update()



import logging
import threading
from concurrent.futures import Future, InvalidStateError
from typing import Dict, Tuple

from scaler.client.agent.mixins import FutureManager
from scaler.client.future import ScalerFuture
from scaler.client.serializer.mixins import Serializer
from scaler.io.utility import concat_list_of_bytes
from scaler.protocol.python.common import TaskStatus
from scaler.protocol.python.message import ObjectResponse, TaskCancel, TaskResult
from scaler.utility.exceptions import DisconnectedError, NoWorkerError, TaskNotFoundError, WorkerDiedError
from scaler.utility.metadata.profile_result import retrieve_profiling_result_from_task_result
from scaler.utility.object_utility import deserialize_failure


class ClientFutureManager(FutureManager):
    def __init__(self, serializer: Serializer):
        self._lock = threading.RLock()
        self._serializer = serializer

        self._task_id_to_future: Dict[bytes, ScalerFuture] = dict()
        self._object_id_to_future: Dict[bytes, Tuple[TaskStatus, ScalerFuture]] = dict()

    def add_future(self, future: Future):
        assert isinstance(future, ScalerFuture)
        with self._lock:
            future.set_running_or_notify_cancel()
            self._task_id_to_future[future.task_id] = future

    def cancel_all_futures(self):
        with self._lock:
            logging.info(f"canceling {len(self._task_id_to_future)} task(s)")
            for task_id, future in self._task_id_to_future.items():
                future.cancel()

            self._task_id_to_future.clear()

    def set_all_futures_with_exception(self, exception: Exception):
        with self._lock:
            for future in self._task_id_to_future.values():
                try:
                    future.set_exception(exception)
                except InvalidStateError:
                    continue  # Future got canceled

            self._task_id_to_future.clear()

    def on_task_result(self, result: TaskResult):
        with self._lock:
            task_id = result.task_id
            if task_id not in self._task_id_to_future:
                return

            future = self._task_id_to_future.pop(task_id)
            assert result.task_id == future.task_id

            profile_result = retrieve_profiling_result_from_task_result(result)

            try:
                if result.status == TaskStatus.NotFound:
                    future.set_exception(TaskNotFoundError(f"task not found: {task_id.hex()}"), profile_result)
                    return

                if result.status == TaskStatus.WorkerDied:
                    future.set_exception(
                        WorkerDiedError(f"worker died when processing task: {task_id.hex()}"), profile_result
                    )
                    return

                if result.status == TaskStatus.NoWorker:
                    future.set_exception(
                        NoWorkerError(f"no available worker when processing task: {task_id.hex()}"), profile_result
                    )
                    return

                if result.status == TaskStatus.Canceled:
                    future.set_exception(DisconnectedError("client disconnected"), profile_result)
                    return

                if result.status == TaskStatus.Success:
                    assert len(result.results) == 1
                    result_object_id = result.results[0]
                    future.set_result_ready(result_object_id, profile_result)
                    self._object_id_to_future[result_object_id] = result.status, future
                    return

                if result.status == TaskStatus.Failed:
                    assert len(result.results) == 1
                    result_object_id = result.results[0]
                    future.set_result_ready(result_object_id, profile_result)
                    self._object_id_to_future[result_object_id] = result.status, future
                    return

                raise TypeError(f"Unknown task status: {result.status}")
            except InvalidStateError:
                return  # Future got canceled

    def on_cancel_task(self, task_cancel: TaskCancel):
        with self._lock:
            self._task_id_to_future.pop(task_cancel.task_id, None)

    def on_object_response(self, response: ObjectResponse):
        for object_id, object_name, object_bytes in zip(
            response.object_content.object_ids,
            response.object_content.object_names,
            response.object_content.object_bytes,
        ):
            if object_id not in self._object_id_to_future:
                continue

            status, future = self._object_id_to_future.pop(object_id)

            try:
                if status == TaskStatus.Success:
                    future.set_result(self._serializer.deserialize(concat_list_of_bytes(object_bytes)))

                elif status == TaskStatus.Failed:
                    future.set_exception(deserialize_failure(concat_list_of_bytes(object_bytes)))
            except InvalidStateError:
                continue  # future got canceled



import asyncio
import logging
import os
import tempfile
import uuid
from typing import Dict, List, Optional, Tuple

import tblib.pickling_support
import zmq.asyncio

# from scaler.utility.logging.utility import setup_logger
from scaler.io.async_binder import AsyncBinder
from scaler.io.async_connector import AsyncConnector
from scaler.io.utility import chunk_to_list_of_bytes
from scaler.protocol.python.common import ObjectContent, TaskStatus
from scaler.protocol.python.message import (
    ObjectInstruction,
    ObjectRequest,
    ObjectResponse,
    ProcessorInitialized,
    Task,
    TaskResult,
)
from scaler.protocol.python.mixins import Message
from scaler.utility.exceptions import ProcessorDiedError
from scaler.utility.metadata.profile_result import ProfileResult
from scaler.utility.mixins import Looper
from scaler.utility.object_utility import generate_object_id, serialize_failure
from scaler.utility.zmq_config import ZMQConfig, ZMQType
from scaler.worker.agent.mixins import HeartbeatManager, ObjectTracker, ProcessorManager, ProfilingManager, TaskManager
from scaler.worker.agent.processor_holder import ProcessorHolder


class VanillaProcessorManager(Looper, ProcessorManager):
    def __init__(
        self,
        context: zmq.asyncio.Context,
        event_loop: str,
        garbage_collect_interval_seconds: int,
        trim_memory_threshold_bytes: int,
        hard_processor_suspend: bool,
        logging_paths: Tuple[str, ...],
        logging_level: str,
    ):
        tblib.pickling_support.install()

        self._event_loop = event_loop

        self._garbage_collect_interval_seconds = garbage_collect_interval_seconds
        self._trim_memory_threshold_bytes = trim_memory_threshold_bytes
        self._hard_processor_suspend = hard_processor_suspend
        self._logging_paths = logging_paths
        self._logging_level = logging_level

        self._address_path = os.path.join(tempfile.gettempdir(), f"scaler_worker_{uuid.uuid4().hex}")
        self._address = ZMQConfig(ZMQType.ipc, host=self._address_path)

        self._heartbeat: Optional[HeartbeatManager] = None
        self._task_manager: Optional[TaskManager] = None
        self._profiling_manager: Optional[ProfilingManager] = None
        self._object_tracker: Optional[ObjectTracker] = None
        self._connector_external: Optional[AsyncConnector] = None

        self._current_holder: Optional[ProcessorHolder] = None
        self._suspended_holders_by_task_id: Dict[bytes, ProcessorHolder] = {}
        self._holders_by_processor_id: Dict[bytes, ProcessorHolder] = {}

        self._can_accept_task_lock: asyncio.Lock = asyncio.Lock()

        self._binder_internal: AsyncBinder = AsyncBinder(
            context=context, name="processor_manager", address=self._address, identity=None
        )
        self._binder_internal.register(self.__on_receive_internal)

    def register(
        self,
        heartbeat: HeartbeatManager,
        task_manager: TaskManager,
        profiling_manager: ProfilingManager,
        object_tracker: ObjectTracker,
        connector_external: AsyncConnector,
    ):
        self._heartbeat = heartbeat
        self._task_manager = task_manager
        self._profiling_manager = profiling_manager
        self._object_tracker = object_tracker
        self._connector_external = connector_external

    async def initialize(self):
        # setup_logger()
        await self._can_accept_task_lock.acquire()  # prevents processor to accept task until initialized
        self.__start_new_processor()

    async def routine(self):
        await self._binder_internal.routine()

    async def on_object_instruction(self, instruction: ObjectInstruction):
        processor_instructions = self._object_tracker.on_object_instruction(instruction)

        for processor_id, instruction in processor_instructions.items():
            await self._binder_internal.send(processor_id, instruction)

    async def on_object_response(self, response: ObjectResponse):
        processors_ids = self._object_tracker.on_object_response(response)

        for process_id in processors_ids:
            await self._binder_internal.send(process_id, response)

    def can_accept_task(self) -> bool:
        return self._can_accept_task_lock.locked()

    async def wait_until_can_accept_task(self):
        """
        Makes sure a processor is ready to start processing a new or suspended task.

        Must be called before any call to `on_task()` or `on_task_resume()`.
        """

        await self._can_accept_task_lock.acquire()

    async def on_task(self, task: Task) -> bool:
        assert self._can_accept_task_lock.locked()
        assert self.initialized()

        holder = self._current_holder

        assert holder.task() is None
        holder.set_task(task)

        self._profiling_manager.on_task_start(holder.pid(), task.task_id)

        await self._binder_internal.send(holder.processor_id(), task)

        return True

    async def on_cancel_task(self, task_id: bytes) -> Optional[Task]:
        assert self._current_holder is not None

        if self.current_task_id() == task_id:
            current_task = self.current_task()
            self.__restart_current_processor(f"cancel task_id={task_id.hex()}")
            return current_task

        if task_id in self._suspended_holders_by_task_id:
            suspended_holder = self._suspended_holders_by_task_id.pop(task_id)
            task = suspended_holder.task()
            self.__kill_processor(f"cancel suspended task_id={task_id.hex()}", suspended_holder)
            return task

        return None

    async def on_failing_processor(self, processor_id: bytes, process_status: str):
        assert self._current_holder is not None

        holder = self._holders_by_processor_id.get(processor_id)

        if holder is None:
            return

        task = holder.task()
        if task is not None:
            profile_result = self.__end_task(holder)  # profiling the task should happen before killing the processor
        else:
            profile_result = None

        reason = f"process died {process_status=}"
        if holder == self._current_holder:
            self.__restart_current_processor(reason)
        else:
            self.__kill_processor(reason, holder)

        if task is not None:
            source = task.source
            task_id = task.task_id

            result_object_bytes = chunk_to_list_of_bytes(serialize_failure(ProcessorDiedError(f"{process_status=}")))

            result_object_id = generate_object_id(source, uuid.uuid4().bytes)
            await self._connector_external.send(
                ObjectInstruction.new_msg(
                    ObjectInstruction.ObjectInstructionType.Create,
                    source,
                    ObjectContent.new_msg(
                        (result_object_id,), (ObjectContent.ObjectContentType.Object,), (b"",), (result_object_bytes,)
                    ),
                )
            )

            await self._task_manager.on_task_result(
                TaskResult.new_msg(task_id, TaskStatus.Failed, profile_result.serialize(), [result_object_id])
            )

    async def on_suspend_task(self, task_id: bytes) -> bool:
        assert self._current_holder is not None
        holder = self._current_holder

        current_task = holder.task()

        if current_task is None or current_task.task_id != task_id:
            return False

        holder.suspend()
        self._suspended_holders_by_task_id[task_id] = holder

        logging.info(f"Worker[{os.getpid()}]: suspend Processor[{holder.pid()}]")

        self.__start_new_processor()

        return True

    def on_resume_task(self, task_id: bytes) -> bool:
        assert self._can_accept_task_lock.locked()
        assert self.initialized()

        if self.current_task() is not None:
            return False

        suspended_holder = self._suspended_holders_by_task_id.pop(task_id, None)

        if suspended_holder is None:
            return False

        self.__kill_processor("replaced by suspended processor", self._current_holder)

        self._current_holder = suspended_holder
        suspended_holder.resume()

        logging.info(f"Worker[{os.getpid()}]: resume Processor[{self._current_holder.pid()}]")

        return True

    def destroy(self, reason: str):
        self.__kill_all_processors(reason)
        self._binder_internal.destroy()
        os.remove(self._address_path)

    def initialized(self) -> bool:
        return self._current_holder is not None and self._current_holder.initialized()

    def current_task(self) -> Optional[Task]:
        assert self._current_holder is not None
        return self._current_holder.task()

    def current_task_id(self) -> bytes:
        task = self.current_task()

        if task is None:
            return b""
        else:
            return task.task_id

    def processors(self) -> List[ProcessorHolder]:
        return list(self._holders_by_processor_id.values())

    def num_suspended_processors(self) -> int:
        return len(self._suspended_holders_by_task_id)

    def __start_new_processor(self):
        self._current_holder = ProcessorHolder(
            self._event_loop,
            self._address,
            self._garbage_collect_interval_seconds,
            self._trim_memory_threshold_bytes,
            self._hard_processor_suspend,
            self._logging_paths,
            self._logging_level,
        )

        processor_pid = self._current_holder.pid()

        self._profiling_manager.on_process_start(processor_pid)

        logging.info(f"Worker[{os.getpid()}]: start Processor[{processor_pid}]")

    def __kill_processor(self, reason: str, holder: ProcessorHolder):
        processor_pid = holder.pid()

        self._profiling_manager.on_process_end(processor_pid)

        if holder.initialized():
            self._holders_by_processor_id.pop(holder.processor_id(), None)
            self._object_tracker.on_processor_end(holder.processor_id())

        holder.kill()

        logging.info(f"Worker[{os.getpid()}]: stop Processor[{processor_pid}], reason: {reason}")

    def __restart_current_processor(self, reason: str):
        assert self._current_holder is not None

        self.__kill_processor(reason, self._current_holder)
        self.__start_new_processor()

    def __kill_all_processors(self, reason: str):
        if self._current_holder is not None:
            self.__kill_processor(reason, self._current_holder)
            self._current_holder = None

        for processor_holder in self._suspended_holders_by_task_id.values():
            self.__kill_processor(reason, processor_holder)

        self._suspended_holders_by_task_id = {}
        self._holders_by_processor_id = {}

    def __end_task(self, processor_holder: ProcessorHolder) -> ProfileResult:
        profile_result = self._profiling_manager.on_task_end(processor_holder.pid(), processor_holder.task().task_id)
        processor_holder.set_task(None)

        return profile_result

    async def __on_receive_internal(self, processor_id: bytes, message: Message):
        if isinstance(message, ProcessorInitialized):
            await self.__on_internal_processor_initialized(processor_id)
            return

        if isinstance(message, ObjectRequest):
            await self.__on_internal_object_request(processor_id, message)
            return

        if isinstance(message, ObjectInstruction):
            await self.__on_internal_object_instruction(processor_id, message)
            return

        if isinstance(message, TaskResult):
            await self.__on_internal_task_result(processor_id, message)
            return

        raise TypeError(f"Unknown message from {processor_id=}: {message}")

    async def __on_internal_processor_initialized(self, processor_id: bytes):
        assert self._current_holder is not None

        if self._current_holder.initialized():
            return

        self._holders_by_processor_id[processor_id] = self._current_holder
        self._current_holder.initialize(processor_id)

        self._can_accept_task_lock.release()

    async def __on_internal_object_request(self, processor_id: bytes, request: ObjectRequest):
        if not self.__processor_ready_to_process_object(processor_id):
            return

        self._object_tracker.on_object_request(processor_id, request)
        await self._connector_external.send(request)

    async def __on_internal_object_instruction(self, processor_id: bytes, instruction: ObjectInstruction):
        if not self.__processor_ready_to_process_object(processor_id):
            return

        await self._connector_external.send(instruction)

    async def __on_internal_task_result(self, processor_id: bytes, task_result: TaskResult):
        assert self._current_holder is not None
        task_id = task_result.task_id

        if task_id == self.current_task_id():
            assert self._current_holder.processor_id() == processor_id

            profile_result = self.__end_task(self._current_holder)

            release_task_lock = True
        elif task_id in self._suspended_holders_by_task_id:
            # Receiving a task result from a suspended processor is possible as the message might have been queued while
            # we were suspending the process.

            holder = self._suspended_holders_by_task_id.pop(task_id)
            assert holder.processor_id() == processor_id

            profile_result = self.__end_task(holder)

            self.__kill_processor("task finished in suspended processor", holder)

            release_task_lock = False
        else:
            return

        await self._task_manager.on_task_result(
            TaskResult.new_msg(
                task_id=task_id,
                status=task_result.status,
                metadata=profile_result.serialize(),
                results=task_result.results,
            )
        )

        # task lock must be released after calling `TaskManager.on_task_result()`
        if release_task_lock:
            self._can_accept_task_lock.release()

    def __processor_ready_to_process_object(self, processor_id: bytes) -> bool:
        holder = self._holders_by_processor_id.get(processor_id)

        if holder is None:
            return False

        assert holder.initialized()

        if holder.task() is None:
            return False

        # TODO: check if the objects belong to the task

        return True



from typing import Optional, Set

from scaler.client.agent.mixins import ObjectManager
from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.common import ObjectContent
from scaler.protocol.python.message import ObjectInstruction, ObjectRequest, TaskResult


class ClientObjectManager(ObjectManager):
    def __init__(self, identity: bytes):
        self._sent_object_ids: Set[bytes] = set()
        self._sent_serializer_id: Optional[bytes] = None

        self._identity = identity

        self._connector_internal: Optional[AsyncConnector] = None
        self._connector_external: Optional[AsyncConnector] = None

    def register(self, connector_internal: AsyncConnector, connector_external: AsyncConnector):
        self._connector_internal = connector_internal
        self._connector_external = connector_external

    async def on_object_instruction(self, instruction: ObjectInstruction):
        if instruction.instruction_type == ObjectInstruction.ObjectInstructionType.Create:
            await self.__send_object_creation(instruction)
        elif instruction.instruction_type == ObjectInstruction.ObjectInstructionType.Delete:
            await self.__delete_objects(instruction)
        elif instruction.instruction_type == ObjectInstruction.ObjectInstructionType.Clear:
            await self.clear_all_objects(clear_serializer=False)

    async def on_object_request(self, object_request: ObjectRequest):
        assert object_request.request_type == ObjectRequest.ObjectRequestType.Get
        await self._connector_external.send(object_request)

    def on_task_result(self, task_result: TaskResult):
        # TODO: received result objects should be deleted from the scheduler when no longer needed.
        # This requires to not delete objects that are required by not-yet-computed dependent graph tasks.
        # For now, we just remove the objects when the client makes a clear request, or on client shutdown.
        # https://github.com/Citi/scaler/issues/43

        self._sent_object_ids.update(task_result.results)

    async def clear_all_objects(self, clear_serializer):
        cleared_object_ids = self._sent_object_ids.copy()

        if clear_serializer:
            self._sent_serializer_id = None
        elif self._sent_serializer_id is not None:
            cleared_object_ids.remove(self._sent_serializer_id)

        self._sent_object_ids.difference_update(cleared_object_ids)

        await self._connector_external.send(
            ObjectInstruction.new_msg(
                ObjectInstruction.ObjectInstructionType.Delete,
                self._identity,
                ObjectContent.new_msg(tuple(cleared_object_ids)),
            )
        )

    async def __send_object_creation(self, instruction: ObjectInstruction):
        assert instruction.instruction_type == ObjectInstruction.ObjectInstructionType.Create

        new_object_ids = set(instruction.object_content.object_ids) - self._sent_object_ids
        if not new_object_ids:
            return

        if ObjectContent.ObjectContentType.Serializer in instruction.object_content.object_types:
            if self._sent_serializer_id is not None:
                raise ValueError("trying to send multiple serializers.")

            serializer_index = instruction.object_content.object_types.index(ObjectContent.ObjectContentType.Serializer)
            self._sent_serializer_id = instruction.object_content.object_ids[serializer_index]

        new_object_content = ObjectContent.new_msg(
            *zip(
                *filter(
                    lambda object_pack: object_pack[0] in new_object_ids,
                    zip(
                        instruction.object_content.object_ids,
                        instruction.object_content.object_types,
                        instruction.object_content.object_names,
                        instruction.object_content.object_bytes,
                    ),
                )
            )
        )

        self._sent_object_ids.update(set(new_object_content.object_ids))

        await self._connector_external.send(
            ObjectInstruction.new_msg(
                ObjectInstruction.ObjectInstructionType.Create, instruction.object_user, new_object_content
            )
        )

    async def __delete_objects(self, instruction: ObjectInstruction):
        assert instruction.instruction_type == ObjectInstruction.ObjectInstructionType.Delete

        if self._sent_serializer_id in instruction.object_content.object_ids:
            raise ValueError("trying to delete serializer.")

        self._sent_object_ids.difference_update(instruction.object_content.object_ids)

        await self._connector_external.send(instruction)



import asyncio
import unittest

from tests.utility import logging_test_name

from scaler.utility.logging.utility import setup_logger
from scaler.utility.queues.async_priority_queue import AsyncPriorityQueue


class TestAsyncPriorityQueue(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)

    def test_async_priority_queue(self):
        async def async_test():
            queue = AsyncPriorityQueue()
            await queue.put((5, 5))
            await queue.put((2, 2))
            await queue.put((6, 6))
            await queue.put((-3, 0))  # supports negative priorities
            await queue.put((1, 1))
            await queue.put((4, 4))
            await queue.put((3, 3))

            queue.remove(2)
            queue.remove(3)
            self.assertEqual(queue.qsize(), 5)

            queue.decrease_priority(4)  # (4, 4) becomes (3, 4)

            self.assertEqual(await queue.get(), (-3, 0))
            self.assertEqual(await queue.get(), (1, 1))
            self.assertEqual(await queue.get(), (3, 4))
            self.assertEqual(await queue.get(), (5, 5))
            self.assertEqual(await queue.get(), (6, 6))
            self.assertEqual(queue.qsize(), 0)
            self.assertTrue(not queue)
            self.assertTrue(queue.empty())

        asyncio.run(async_test())



import os
import time
import unittest

from tests.utility import get_available_tcp_port, logging_test_name

from scaler import Client, Cluster, SchedulerClusterCombo
from scaler.utility.logging.utility import setup_logger


def sleep_and_return_pid(sec: int):
    time.sleep(sec)
    return os.getpid()


class TestBalance(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)

    def test_balance(self):
        """
        Schedules a few long-lasting tasks to a single process cluster, then adds workers. We expect the remaining tasks
        to be balanced to the new workers.
        """

        N_TASKS = 8
        N_WORKERS = N_TASKS

        address = f"tcp://127.0.0.1:{get_available_tcp_port()}"
        combo = SchedulerClusterCombo(
            address=address,
            n_workers=1,
            per_worker_queue_size=N_TASKS,
            load_balance_seconds=1,  # FIXME: re-enable balancing as it's currently disabled by default
        )

        client = Client(address=address)

        futures = [client.submit(sleep_and_return_pid, 10) for _ in range(N_TASKS)]

        time.sleep(5)

        new_cluster = Cluster(
            address=combo._cluster._address,
            worker_io_threads=1,
            worker_names=[str(i) for i in range(0, N_WORKERS - 1)],
            heartbeat_interval_seconds=combo._cluster._heartbeat_interval_seconds,
            task_timeout_seconds=combo._cluster._task_timeout_seconds,
            death_timeout_seconds=combo._cluster._death_timeout_seconds,
            garbage_collect_interval_seconds=combo._cluster._garbage_collect_interval_seconds,
            trim_memory_threshold_bytes=combo._cluster._trim_memory_threshold_bytes,
            hard_processor_suspend=combo._cluster._hard_processor_suspend,
            event_loop=combo._cluster._event_loop,
            logging_paths=combo._cluster._logging_paths,
            logging_level=combo._cluster._logging_level,
            logging_config_file=combo._cluster._logging_config_file,
        )
        new_cluster.start()

        pids = {f.result() for f in futures}

        self.assertEqual(len(pids), N_WORKERS)

        client.disconnect()

        new_cluster.terminate()
        combo.shutdown()



from asyncio import Queue, QueueEmpty
from typing import Hashable

from scaler.utility.queues.indexed_queue import IndexedQueue


class AsyncIndexedQueue(Queue):
    """This should have same set of features as asyncio.Queue, with additional methods like remove
    - it behaves like regular async queue, except:
      - all the items pushed to queue should be hashable
      - those items should be unique in queue
    - IndexedQueue.put(), IndexedQueue.get(), IndexedQueue.remove() should all take O(1) time complexity
    """

    def __contains__(self, item):
        return item in self._queue

    def __len__(self):
        return self._queue.__len__()

    def _init(self, maxsize):
        self._queue = IndexedQueue()

    def _put(self, item: Hashable):
        self._queue.put(item)

    def _get(self):
        try:
            return self._queue.get()
        except IndexError:
            raise QueueEmpty(f"{self.__class__.__name__} queue empty")

    def remove(self, item: Hashable):
        """remove the item in the queue in O(1) time complexity"""
        self._queue.remove(item)



import abc


class Looper(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def routine(self):
        raise NotImplementedError()


class Reporter(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_status(self):
        raise NotImplementedError()



from typing import Dict, Generic, Iterable, Set, Tuple, TypeVar

KeyT = TypeVar("KeyT")
ValueT = TypeVar("ValueT")

LeftKeyT = TypeVar("LeftKeyT")
RightKeyT = TypeVar("RightKeyT")


class ManyToManyDict(Generic[LeftKeyT, RightKeyT]):
    def __init__(self):
        self._left_key_to_right_key_set: _KeyValueDictSet[LeftKeyT, RightKeyT] = _KeyValueDictSet()
        self._right_key_to_left_key_set: _KeyValueDictSet[RightKeyT, LeftKeyT] = _KeyValueDictSet()

    def left_keys(self):
        return self._left_key_to_right_key_set.keys()

    def right_keys(self):
        return self._right_key_to_left_key_set.keys()

    def add(self, left_key: LeftKeyT, right_key: RightKeyT):
        self._left_key_to_right_key_set.add(left_key, right_key)
        self._right_key_to_left_key_set.add(right_key, left_key)

    def remove(self, left_key: LeftKeyT, right_key: RightKeyT):
        self._left_key_to_right_key_set.remove_value(left_key, right_key)
        self._right_key_to_left_key_set.remove_value(right_key, left_key)

    def has_left_key(self, left_key: LeftKeyT) -> bool:
        return left_key in self._left_key_to_right_key_set

    def has_right_key(self, right_key: RightKeyT) -> bool:
        return right_key in self._right_key_to_left_key_set

    def has_key_pair(self, left_key: LeftKeyT, right_key: RightKeyT) -> bool:
        return (
            self.has_left_key(left_key)
            and self.has_right_key(right_key)
            and right_key in self._left_key_to_right_key_set.get_values(left_key)
            and left_key in self._right_key_to_left_key_set.get_values(right_key)
        )

    def left_key_items(self) -> Iterable[Tuple[LeftKeyT, Set[RightKeyT]]]:
        return self._left_key_to_right_key_set.items()

    def right_key_items(self) -> Iterable[Tuple[RightKeyT, Set[LeftKeyT]]]:
        return self._right_key_to_left_key_set.items()

    def get_left_items(self, right_key: RightKeyT) -> Set[LeftKeyT]:
        if right_key not in self._right_key_to_left_key_set:
            raise ValueError(f"cannot find {right_key=} in ManyToManyDict")

        return self._right_key_to_left_key_set.get_values(right_key)

    def get_right_items(self, left_key: LeftKeyT) -> Set[RightKeyT]:
        if left_key not in self._left_key_to_right_key_set:
            raise ValueError(f"cannot find {left_key=} in ManyToManyDict")

        return self._left_key_to_right_key_set.get_values(left_key)

    def remove_left_key(self, left_key: LeftKeyT) -> Set[RightKeyT]:
        if left_key not in self._left_key_to_right_key_set:
            raise KeyError(f"cannot find {left_key=} in ManyToManyDict")

        right_keys = self._left_key_to_right_key_set.remove_key(left_key)
        for right_key in right_keys:
            self._right_key_to_left_key_set.remove_value(right_key, left_key)

        return right_keys

    def remove_right_key(self, right_key: RightKeyT) -> Set[LeftKeyT]:
        if right_key not in self._right_key_to_left_key_set:
            raise ValueError(f"cannot find {right_key=} in ManyToManyDict")

        left_keys = self._right_key_to_left_key_set.remove_key(right_key)
        for left_key in left_keys:
            self._left_key_to_right_key_set.remove_value(left_key, right_key)

        return left_keys


class _KeyValueDictSet(Generic[KeyT, ValueT]):
    def __init__(self):
        self._key_to_value_set: Dict[KeyT, Set[ValueT]] = dict()

    def __contains__(self, key) -> bool:
        return key in self._key_to_value_set

    def keys(self):
        return self._key_to_value_set.keys()

    def values(self):
        return self._key_to_value_set.values()

    def items(self):
        return self._key_to_value_set.items()

    def add(self, key: KeyT, value: ValueT):
        if key not in self._key_to_value_set:
            self._key_to_value_set[key] = set()

        self._key_to_value_set[key].add(value)

    def get_values(self, key: KeyT) -> Set[ValueT]:
        if key not in self._key_to_value_set:
            raise ValueError(f"cannot find {key=} in KeyValueSet")

        return self._key_to_value_set[key]

    def remove_key(self, key: KeyT) -> Set[ValueT]:
        if key not in self._key_to_value_set:
            raise KeyError(f"cannot find {key=} in KeyValueSet")

        values = self._key_to_value_set.pop(key)
        return values

    def remove_value(self, key: KeyT, value: ValueT):
        if key not in self._key_to_value_set:
            raise KeyError(f"cannot find {key=} in KeyValueSet")

        self._key_to_value_set[key].remove(value)
        if not self._key_to_value_set[key]:
            self._key_to_value_set.pop(key)



import abc
from typing import Callable, Dict, Generator, Generic, Optional, Set, Tuple, TypeVar

from scaler.utility.many_to_many_dict import ManyToManyDict

ObjectKeyType = TypeVar("ObjectKeyType")


class ObjectUsage(Generic[ObjectKeyType], metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get_object_key(self) -> ObjectKeyType:
        raise NotImplementedError()


ObjectType = TypeVar("ObjectType", bound=ObjectUsage)


class ObjectTracker(Generic[ObjectKeyType, ObjectType]):
    def __init__(self, prefix: str, callback: Callable[[ObjectType], None]):
        self._prefix = prefix
        self._callback = callback

        self._current_blocks: Set[ObjectKeyType] = set()
        self._object_key_to_block: ManyToManyDict[ObjectKeyType, ObjectKeyType] = ManyToManyDict()
        self._object_key_to_object: Dict[ObjectKeyType, ObjectType] = dict()

    def object_count(self):
        return len(self._object_key_to_object)

    def items(self):
        return self._object_key_to_object.items()

    def get_all_object_keys(self) -> Set[ObjectKeyType]:
        return set(self._object_key_to_object.keys())

    def has_object(self, key: ObjectKeyType) -> bool:
        return key in self._object_key_to_object

    def get_object(self, key: ObjectKeyType) -> ObjectType:
        return self._object_key_to_object[key]

    def add_object(self, obj: ObjectType):
        self._object_key_to_object[obj.get_object_key()] = obj

    def get_object_block_pairs(
        self, blocks: Set[ObjectKeyType]
    ) -> Generator[Tuple[ObjectKeyType, ObjectKeyType], None, None]:
        for block in blocks:
            if not self._object_key_to_block.has_right_key(block):
                continue

            for object_key in self._object_key_to_block.get_left_items(block):
                yield object_key, block

    def add_blocks_for_one_object(self, object_key: ObjectKeyType, blocks: Set[ObjectKeyType]):
        if object_key not in self._object_key_to_object:
            raise KeyError(f"cannot find key={object_key} in ObjectTracker")

        for block in blocks:
            self._object_key_to_block.add(object_key, block)

        self._current_blocks.update(blocks)

    def remove_blocks_for_one_object(self, object_key: ObjectKeyType, blocks: Set[ObjectKeyType]):
        ready_objects = []
        for block in blocks:
            obj = self.__remove_block_for_object(object_key, block)
            if obj is None:
                continue

            ready_objects.append(obj)

        for obj in ready_objects:
            self._callback(obj)

    def add_one_block_for_objects(self, object_keys: Set[ObjectKeyType], block: ObjectKeyType):
        for object_key in object_keys:
            if object_key not in self._object_key_to_object:
                raise KeyError(f"cannot find key={object_key} in ObjectTracker")

            self._object_key_to_block.add(object_key, block)

        self._current_blocks.add(block)

    def remove_one_block_for_objects(self, object_keys: Set[ObjectKeyType], block: ObjectKeyType):
        ready_objects = []
        for object_key in object_keys:
            obj = self.__remove_block_for_object(object_key, block)
            if obj is None:
                continue

            ready_objects.append(obj)

        for obj in ready_objects:
            self._callback(obj)

    def remove_blocks(self, blocks: Set[ObjectKeyType]):
        ready_objects = []
        for block in blocks:
            if not self._object_key_to_block.has_right_key(block):
                continue

            object_keys = self._object_key_to_block.get_left_items(block).copy()
            for object_key in object_keys:
                obj = self.__remove_block_for_object(object_key, block)
                if obj is None:
                    continue

                ready_objects.append(obj)

        for obj in ready_objects:
            self._callback(obj)

    def __remove_block_for_object(self, object_key: ObjectKeyType, block: ObjectKeyType) -> Optional[ObjectType]:
        if block not in self._current_blocks:
            return None

        if object_key not in self._object_key_to_object:
            return None

        if not self._object_key_to_block.has_key_pair(object_key, block):
            return None

        self._object_key_to_block.remove(object_key, block)

        if not self._object_key_to_block.has_right_key(block):
            self._current_blocks.remove(block)

        if self._object_key_to_block.has_left_key(object_key):
            return None

        return self._object_key_to_object.pop(object_key)



import dataclasses
from collections import deque
from threading import Lock
from typing import Deque

from nicegui import ui

from scaler.protocol.python.common import TaskStatus
from scaler.protocol.python.message import StateTask
from scaler.utility.formatter import format_bytes
from scaler.utility.metadata.profile_result import ProfileResult

# TaskStatus values corresponding to completed tasks (some are in-progress e.g. Running)
COMPLETED_TASK_STATUSES = {
    TaskStatus.Success,
    TaskStatus.Failed,
    TaskStatus.Canceled,
    TaskStatus.WorkerDied,
    TaskStatus.NoWorker,
}


@dataclasses.dataclass
class TaskData:
    task: str = dataclasses.field(default="")
    function: str = dataclasses.field(default="")
    duration: str = dataclasses.field(default="")
    peak_mem: str = dataclasses.field(default="")
    status: str = dataclasses.field(default="")

    def populate(self, state: StateTask):
        self.task = f"{state.task_id.hex()}"
        self.function = state.function_name.decode()
        self.status = state.status.name

        self.duration = "N/A"
        self.peak_mem = "N/A"
        if state.metadata:
            profiling_data = ProfileResult.deserialize(state.metadata)
            duration = profiling_data.duration_s
            mem = profiling_data.memory_peak
            self.duration = f"{duration:.2f}s"
            self.peak_mem = format_bytes(mem) if mem != 0 else "0"

    def draw_row(self):
        color = "color: green" if self.status == "Success" else "color: red"
        ui.label(self.task)
        ui.label(self.function)
        ui.label(self.duration)
        ui.label(self.peak_mem)
        ui.label(self.status).style(color)

    @staticmethod
    def draw_titles():
        ui.label("Task ID")
        ui.label("Function")
        ui.label("Duration")
        ui.label("Peak mem")
        ui.label("Status")


class TaskLogTable:
    def __init__(self):
        self._task_log: Deque[TaskData] = deque(maxlen=100)
        self._lock = Lock()

    def handle_task_state(self, state: StateTask):
        if state.status not in COMPLETED_TASK_STATUSES:
            return

        row = TaskData()
        row.populate(state)

        with self._lock:
            self._task_log.appendleft(row)

    @ui.refreshable
    def draw_section(self):
        with self._lock:
            with ui.card().classes("w-full q-mx-auto"), ui.grid(columns=5).classes("q-mx-auto"):
                TaskData.draw_titles()
                for task in self._task_log:
                    task.draw_row()



import threading
from concurrent.futures import Future, InvalidStateError
from typing import Any, Callable, Optional

from scaler.io.sync_connector import SyncConnector
from scaler.protocol.python.message import ObjectRequest, Task, TaskCancel
from scaler.utility.event_list import EventList
from scaler.utility.metadata.profile_result import ProfileResult


class ScalerFuture(Future):
    def __init__(self, task: Task, is_delayed: bool, group_task_id: Optional[bytes], connector: SyncConnector):
        super().__init__()

        self._waiters = EventList(self._waiters)  # type: ignore[assignment]
        self._waiters.add_update_callback(self._on_waiters_updated)  # type: ignore[attr-defined]

        self._task_id: bytes = task.task_id
        self._is_delayed: bool = is_delayed
        self._group_task_id: Optional[bytes] = group_task_id
        self._connector: SyncConnector = connector

        self._result_object_id: Optional[bytes] = None
        self._result_ready_event = threading.Event()
        self._result_request_sent = False
        self._result_received = False

        self._profiling_info: Optional[ProfileResult] = None

    @property
    def task_id(self):
        return self._task_id

    def profiling_info(self) -> ProfileResult:
        with self._condition:  # type: ignore[attr-defined]
            if self._profiling_info is None:
                raise ValueError(f"didn't receive profiling info for {self} yet")

            return self._profiling_info

    def set_result_ready(self, object_id: Optional[bytes], profile_result: Optional[ProfileResult] = None) -> None:
        with self._condition:  # type: ignore[attr-defined]
            if self.done():
                raise InvalidStateError(f"invalid future state: {self._state}")

            self._state = "FINISHED"

            if object_id is not None:
                self._result_object_id = object_id

            if profile_result is not None:
                self._profiling_info = profile_result

            # if it's not delayed future, or if there is any listener (waiter or callback), get the result immediately
            if not self._is_delayed or self._has_result_listeners():
                self._request_result_object()

            self._result_ready_event.set()

    def _set_result_or_exception(
        self,
        result: Optional[Any] = None,
        exception: Optional[BaseException] = None,
        profiling_info: Optional[ProfileResult] = None,
    ) -> None:
        with self._condition:  # type: ignore[attr-defined]
            if self.cancelled():
                raise InvalidStateError(f"invalid future state: {self._state}")

            if self._result_received:
                raise InvalidStateError("future already received object data.")

            if profiling_info is not None:
                if self._profiling_info is not None:
                    raise InvalidStateError("cannot set profiling info twice.")

                self._profiling_info = profiling_info

            self._state = "FINISHED"
            self._result_received = True

            if exception is not None:
                assert result is None
                self._exception = exception
                for waiter in self._waiters:
                    waiter.add_exception(self)
            else:
                self._result = result
                for waiter in self._waiters:
                    waiter.add_result(self)

            self._result_ready_event.set()
            self._condition.notify_all()

        self._invoke_callbacks()  # type: ignore[attr-defined]

    def set_result(self, result: Any, profiling_info: Optional[ProfileResult] = None) -> None:
        self._set_result_or_exception(result=result, profiling_info=profiling_info)

    def set_exception(self, exception: Optional[BaseException], profiling_info: Optional[ProfileResult] = None) -> None:
        self._set_result_or_exception(exception=exception, profiling_info=profiling_info)

    def result(self, timeout: Optional[float] = None) -> Any:
        self._result_ready_event.wait(timeout)

        with self._condition:  # type: ignore[attr-defined]
            # if it's delayed future, get the result when future.result() gets called
            if self._is_delayed:
                self._request_result_object()

            if not self._result_received:
                self._condition.wait(timeout)

            return super().result()

    def exception(self, timeout: Optional[float] = None) -> Optional[BaseException]:
        self._result_ready_event.wait(timeout)

        with self._condition:  # type: ignore[attr-defined]
            # if it's delayed future, get the result when future.exception() gets called
            if self._is_delayed:
                self._request_result_object()

            if not self._result_received:
                self._condition.wait(timeout)

            return super().exception()

    def cancel(self) -> bool:
        with self._condition:  # type: ignore[attr-defined]
            if self.cancelled():
                return True

            if self.done():
                return False

            if self._group_task_id is not None:
                self._connector.send(TaskCancel.new_msg(self._group_task_id))
            else:
                self._connector.send(TaskCancel.new_msg(self._task_id))

            self._state = "CANCELLED"
            self._result_received = True

            self._result_ready_event.set()
            self._condition.notify_all()  # type: ignore[attr-defined]

        self._invoke_callbacks()  # type: ignore[attr-defined]
        return True

    def add_done_callback(self, fn: Callable[[Future], Any]) -> None:
        with self._condition:  # type: ignore[attr-defined]
            # if it's delayed future, get the result when a callback gets added
            if self._is_delayed:
                self._request_result_object()

            return super().add_done_callback(fn)

    def _on_waiters_updated(self, waiters: EventList):
        with self._condition:  # type: ignore[attr-defined]
            # if it's delayed future, get the result when waiter gets added
            if self._is_delayed and len(self._waiters) > 0:
                self._request_result_object()

    def _has_result_listeners(self) -> bool:
        return len(self._done_callbacks) > 0 or len(self._waiters) > 0  # type: ignore[attr-defined]

    def _request_result_object(self):
        if self._result_request_sent or self._result_object_id is None or self.cancelled():
            return

        self._connector.send(ObjectRequest.new_msg(ObjectRequest.ObjectRequestType.Get, (self._result_object_id,)))
        self._result_request_sent = True



import dataclasses
import pickle
from typing import Any, Callable, List, Optional, Set

import cloudpickle

from scaler.client.serializer.mixins import Serializer
from scaler.io.sync_connector import SyncConnector
from scaler.io.utility import chunk_to_list_of_bytes
from scaler.protocol.python.common import ObjectContent
from scaler.protocol.python.message import ObjectInstruction
from scaler.utility.object_utility import generate_object_id, generate_serializer_object_id


@dataclasses.dataclass
class ObjectCache:
    object_id: bytes
    object_type: ObjectContent.ObjectContentType
    object_name: bytes
    object_bytes: List[bytes]


class ObjectBuffer:
    def __init__(self, identity: bytes, serializer: Serializer, connector: SyncConnector):
        self._identity = identity
        self._serializer = serializer
        self._connector = connector

        self._pending_objects: List[ObjectCache] = list()
        self._pending_delete_objects: Set[bytes] = set()

    def buffer_send_serializer(self) -> ObjectCache:
        serializer_cache = self.__construct_serializer()

        self._pending_objects.append(serializer_cache)
        return serializer_cache

    def buffer_send_function(self, fn: Callable) -> ObjectCache:
        function_cache = self.__construct_function(fn)

        self._pending_objects.append(function_cache)
        return function_cache

    def buffer_send_object(self, obj: Any, name: Optional[str] = None) -> ObjectCache:
        object_cache = self.__construct_object(obj, name)

        self._pending_objects.append(object_cache)
        return object_cache

    def buffer_delete_object(self, object_ids: Set[bytes]):
        self._pending_delete_objects.update(object_ids)

    def commit_send_objects(self):
        if not self._pending_objects:
            return

        objects_to_send = [
            (obj_cache.object_id, obj_cache.object_type, obj_cache.object_name, obj_cache.object_bytes)
            for obj_cache in self._pending_objects
        ]

        self._connector.send(
            ObjectInstruction.new_msg(
                ObjectInstruction.ObjectInstructionType.Create,
                self._identity,
                ObjectContent.new_msg(*zip(*objects_to_send)),
            )
        )

        self._pending_objects.clear()

    def commit_delete_objects(self):
        if not self._pending_delete_objects:
            return

        self._connector.send(
            ObjectInstruction.new_msg(
                ObjectInstruction.ObjectInstructionType.Delete,
                self._identity,
                ObjectContent.new_msg(tuple(self._pending_delete_objects)),
            )
        )

        self._pending_delete_objects.clear()

    def clear(self):
        """
        remove all committed and pending objects.
        """

        self._pending_delete_objects.clear()
        self._pending_objects.clear()

        self._connector.send(
            ObjectInstruction.new_msg(
                ObjectInstruction.ObjectInstructionType.Clear, self._identity, ObjectContent.new_msg(tuple())
            )
        )

    def __construct_serializer(self) -> ObjectCache:
        serializer_bytes = cloudpickle.dumps(self._serializer, protocol=pickle.HIGHEST_PROTOCOL)
        object_id = generate_serializer_object_id(self._identity)
        return ObjectCache(
            object_id,
            ObjectContent.ObjectContentType.Serializer,
            b"serializer",
            chunk_to_list_of_bytes(serializer_bytes),
        )

    def __construct_function(self, fn: Callable) -> ObjectCache:
        function_bytes = self._serializer.serialize(fn)
        object_id = generate_object_id(self._identity, function_bytes)
        function_cache = ObjectCache(
            object_id,
            ObjectContent.ObjectContentType.Object,
            getattr(fn, "__name__", f"<func {object_id.hex()[:6]}>").encode(),
            chunk_to_list_of_bytes(function_bytes),
        )
        return function_cache

    def __construct_object(self, obj: Any, name: Optional[str] = None) -> ObjectCache:
        object_payload = self._serializer.serialize(obj)
        object_id = generate_object_id(self._identity, object_payload)
        name_bytes = name.encode() if name else f"<obj {object_id.hex()[-6:]}>".encode()
        return ObjectCache(
            object_id, ObjectContent.ObjectContentType.Object, name_bytes, chunk_to_list_of_bytes(object_payload)
        )



STORAGE_SIZE_MODULUS = 1024.0
TIME_MODULUS = 1000


def format_bytes(number) -> str:
    for unit in ["B", "K", "M", "G", "T"]:
        if number >= STORAGE_SIZE_MODULUS:
            number /= STORAGE_SIZE_MODULUS
            continue

        if unit in {"B", "K"}:
            return f"{int(number)}{unit}"

        return f"{number:.1f}{unit}"

    raise ValueError("This should not happen")


def format_integer(number):
    return f"{number:,}"


def format_percentage(number: int):
    return f"{(number/1000):.1%}"


def format_microseconds(number: int):
    for unit in ["us", "ms", "s"]:
        if number >= TIME_MODULUS:
            number = int(number / TIME_MODULUS)
            continue

        if unit == "us":
            return f"{number/TIME_MODULUS:.1f}ms"

        too_big_sign = "+" if unit == "s" and number > TIME_MODULUS else ""
        return f"{int(number)}{too_big_sign}{unit}"


def format_seconds(number: int):
    if number > 60:
        return "60+s"

    return f"{number}s"
